; z80dasm 1.1.0
; command line: z80dasm --origin=0x0 --labels --block-def=mms_rom_84a.blk --source --address mms_rom_84a

	include	'z89.inc'
	include 'h17_rom.inc'

; Device IDs - todo: move to central file?
z17_device:		equ	0	; 0,1,2
z17_device_num:		equ	3
z67_device:		equ	3	; 3,4
z67_device_num:		equ	2
z47_device:		equ	5	; 5,6,7,8
z47_device_num:		equ	4
m314_z47_device:	equ	5	; 5,6,7,8 + m314_z47_dev_off
m314_z47_device_num:	equ	4
					; 9,10,11,12,13,14
m314_imi_device:	equ	15	; 15,16,17,18,19,20,21,22,23
m314_imi_device_num:	equ	9
					; 24,25,26,27,28
m316_device:		equ	29	; 29,30,31,32,33,34,35,36
m316_device_num:	equ	8
m316_8_device:		equ	m316_device+0
m316_5_device:		equ	m316_device+4
x37_device:		equ	37	; 37
x37_device_num:		equ	1
x38_device:		equ	38	; 38,39
x38_device_num:		equ	2
					; 40,41,42,43,44,45
z37_device:		equ	46	; 46,47,48,49
z37_device_num:		equ	4
m314_z47_dev_off:	equ	200	; keep 77314 and z47 separate...
sub_device:		equ	0feh
invalid_device:		equ	0ffh

tmp_buf:		equ	boot_loc

memtest_loc:	equ	03000h

; Start ROM code...
	org	00000h
rst00:	jp start_rom	; ROM image had nop,start_rom - must have been copy error

boot_msg:	defb	'oot ',0	; [B]oot

; clk_intr - RTC interrupt
rst08:	call intr_setup
	ld hl,(tic_ctr)
	jp tic_intr

rst10:	equ	$-1	; nop... convenient overflow of rst08 code...
	call intr_setup
	ld a,(de)
	jp rst10_impl

rst18:	jp	rst18_vec

go_msg:	defb	'o ',0
	defs	2,0

rst20:	jp	rst20_vec
	defs	5,0

rst28:	jp	rst28_vec

delay:	jp	timeout

q_prompt:	defb	'?',0

rst30:	jp	rst30_vec
	defs	5,0

rst38:	jp	rst38_vec

subs_msg:
	defb	'ubstitute ',0	; [S]ubstitute
pc_msg:
	defb	'rogram Counter ',0	; [P]rogram Counter

z80_nmi_vec:	equ	0066h	; needs to be in some inc file!

memtest_msg:
	defb	'emory test',0

	defs	z80_nmi_vec-$,0

nmi:	; Z80 NMI vector! must be at 0066h!
	ex (sp),hl	; return address... next? instruction
	push hl
	push af
	dec hl	; see what we were doing when NMI struck
	ld a,(hl)
	cp h8_fp0
	jr z,handle_h8_fp0
	cp h8_fp1
	jr z,handle_h8_noop
	cp h8_con0
	jr z,handle_h8_noop
	cp h8_con1
	jr nz,ret_nothing
; ignore output, return 00 for input
handle_h8_noop:
	dec hl
	ld a,(hl)
	cp 0d3h		; OUT
	jr z,ret_nothing
	cp 0dbh		; IN
	jr nz,ret_nothing
	; did "in a,(nmi_from_[123])"
	pop af		; IN, must be FDC?
	ld a,000h	; return 00
	jr ret_value

handle_h8_fp0:
	dec hl
	ld a,(hl)
	cp 0dbh		; IN
	jr nz,h8fp_not_in
	; did "in a,(h8_fp0)"
	pop af
	ld a,0ffh	; return FF - don't simulate H8 dipswitches?
	jr ret_value

h8fp_not_in:
	cp 0d3h		; OUT
	jr nz,ret_nothing
	; did "out (h8_fp0),a"
	; move bit 6 to gp_2msen, and ~bit 4 to gp_ssen
	pop af	; restore saved A
	push af	; C  AAAAAAAAAAAAAAA
	rla	; 7  6 5 4 3 2 1 0 C
	rla	; 6  5 4 3 2 1 0 C 7
	cpl	; 6  5'4'3'2'1'0'C'7'
	rla	; 5' 4'3'2'1'0'C'7'6
	rlca	; 4' 3'2'1'0'C'7'6 4'
	and gp_2msen|gp_ssen
	ld hl,saved_gp
	or (hl)
	out (gp_port),a
ret_nothing:
	pop af
ret_value:
	pop hl
	ex (sp),hl
	retn

intr_exit:
	ld a,(bc)
	rrca
	call c,rst08_vec
; pop 5, ei, ret...
intr_ret:
	pop af
	pop af
	pop bc
	pop de
	pop hl
ei_ret:
	ei
	ret

tic_intr:
	inc hl
	ld (tic_ctr),hl
	ld bc,saved_h8fp
	ld a,(bc)
	out (h8_fp0),a
	and h8_2msdis
	jr nz,intr_ret	; tic disabled
	dec bc
	ld a,(bc)
	rla
	jr c,intr_exit
	ld hl,10
	add hl,sp
	ld e,(hl)
	inc hl
	ld d,(hl)	; DE = SP[10:11]
	dec de
	ld a,(de)
	cp 076h	; HALT?
	jr nz,intr_exit
	call con_bell
	ld a,'H'
	call con_out
cmd_loop:
	ld hl,saved_h8fp
	ld (hl),h8_bit7|h8_bit6|h8_2msdis|h8_ssen
	ld hl,(start_sp)	; reset SP to origin
	ld sp,hl
	call con_bell
rom_prompt:
	ei
	ld hl,rom_prompt	; restart here on error
	push hl
	ld hl,mms_prompt
	call msgz_out
inval_cmd:
	call con_input2
	and 05fh	; toupper()
	ld hl,mms_cmds
	ld b,num_mms_cmds
search_cmd:
	cp (hl)
	inc hl
	jr z,found_cmd
	inc hl
	inc hl
	djnz search_cmd
	; command not found...
	ld c,a
	ld a,(boot_hook)
	cp 076h	; HALT instruction
	call z,rom_cmd_hook
	call con_bell
	jr inval_cmd

found_cmd:
	call con_out	; echo command char
	ld a,(hl)
	inc hl
	ld h,(hl)
	ld l,a
	jp (hl)

mms_cmds:
	defb	'G'
	defw	go_cmd
	defb	'S'
	defw	subs_cmd
	defb	'P'
	defw	pc_cmd
	defb	'B'
	defw	mms_boot
	defb	'M'
	defw	mms_memtest
num_mms_cmds:	equ	($-mms_cmds)/3

	defs 8,0
l0137h:
	jp z47_wait_tr_in_p1

; [P]rogram Counter command -
pc_cmd:
	ld hl,pc_msg
	call msgz_out
	ld hl,12
	add hl,sp
	ld e,(hl)
	inc hl
	ld d,(hl)	; DE = SP[12:13]
	ex de,hl	; current PC addr into HL, DE = &PC
	call get_hex_wait
	jr c,pc_new_pc	; got a hex digit...
	; must have been CR, display current value...
	call nl_addr_sp_out
	call get_hex_wait
	ret nc
	; CY set, we have first hex digit in A...
pc_new_pc:
	ex de,hl	; PC addr into HL
	ld d,00dh	; term on CR
	jp subs_one_hex	; get hex number and store

go_cmd:
	ld hl,go_msg
	call msgz_out
	ld hl,13
	add hl,sp	; HL = dest addr?
	call get_hex_wait	; get CR or hex digit
	ld d,00dh	; CR
	call c,subs_one_hex
	call con_crlf
	ld a,h8_bit7|h8_bit6|h8_ssen	; !h8_2msdis
	jr go_ret1

single_step:
	di
	ld a,(saved_h8fp)
	xor h8_ssen
	out (h8_fp0),a
go_ret1:
	ld (saved_h8fp),a
	pop hl		; adjust stack...
	jp intr_ret

rst10_impl:
	or 010h
	out (h8_fp0),a	; cause NMI?
	ld (de),a
	and 020h
	jp nz,rom_prompt; abort, restart prompt loop
	jp rst10_vec	; continue with any user-coded rst10 actions

set_timeout_5:
	ld a,5		; boot timeout 5 "meta" tics
set_timeout:
	ld hl,boot_timeout
	ld (rst08_vec+1),hl
	ld (boot_time),a
	ei
	ret

boot_timeout:
	ld hl,tic_ctr
	xor a
	or (hl)
	ret nz
	inc hl
	ld a,(hl)
	rrca
	ret c	; every 512 tics...
	ld hl,boot_time
	dec (hl)
	ret nz
boot_err:
	ld hl,q_prompt
	call msgz_out
	ld hl,ei_ret
	ld (rst08_vec+1),hl
	ld (saved_z17p3),a
	out (z17_port3),a
	jp cmd_loop

auto_boot:
	ld hl,auto_disable
	in a,(gp_port)
	ld d,a
	xor gp_autoboot ;
	or (hl)
	ret m		; return if no auto boot
	; try to auto boot...
	ld (hl),d	; save flags - disable auto boot if we come back here
	call get_def_dev
	ld hl,auto_boot_msg
	call msgz_out
	ld sp,boot_stack
	jp boot_wo_cr

; BLOCK 'data10' (start 0x01d9 end 0x01e3)
auto_boot_msg:
	defb	'Auto Boot',0

; BLOCK 'text10' (start 0x01e3 end 0x04eb)
chk_pri_dev:
	ld d,z17_device
	in a,(gp_port)
chk_dev:
	and gp_pri_dev
	ret z		; gp_pri_z17
	cp z89_pri_z47
	ld d,z47_device
	ret z
	cp z89_pri_z67
	ld d,z67_device
	ret z
	ld d,invalid_device
	ret		; gp_pri_ndi

chk_sec_dev:
	ld d,z37_device
	in a,(gp_port)
	rrca
	rrca		; bit field gp_sec_dev... same as gp_pri_dev except 00 case
	jr chk_dev

; get default boot device as specified by various dip switches
get_def_dev:
	ld de,(z17_device<<8)|0	; unit 0 on Z17 controller is default
	in a,(gp_port)
	and gp_def_boot	; re-assign bits 6-4 to default boot selection
	cp gp_boot_pri
	jr z,chk_pri_dev	; boot according to "primary device" switches
	cp gp_boot_sec
	jr z,chk_sec_dev	; boot according to "secondary device" switches
	ld a,(boot_hook)
	cp 076h	; halt instruction
	ret nz
	jp def_dev_hook	; look for MMS devices

; figure out where the H47 is jumpered...
; C = target "device" - dipsw code 00, 01, 10, 11...
; returns H47 base port in B, CY set, ZR if device present
get_pri_base:
	ld b,z89_pri_base
	in a,(gp_port)
	and gp_pri_dev
	cp c
	ret z
	ld b,z89_sec_base
	in a,(gp_port)
	rrca		;
	rrca		;
	and 003h	; i.e. gp_sec_dev
	cp c
	ret

; This supports 3 drives, unlike the H17 ROM
l_h17_sdp:
	ld a,10	; error count
	ld (h17_oecnt),a
	ld a,(boot_unit);  0, 1, 2
	push af
	add a,-2	; -2,-1, 0
	adc a,3		;  1, 2, 4
	jp h17_sdp_d+10

z47_in_p0:
	or a	; CY = 0: z47_base + 0
z47_in_px:
	push bc
	ld a,(boot_base_port)
	adc a,0		; z47_base + 0 or z47_base + 1
	ld c,a
	in a,(c)
	pop bc
	ret

	defs 27,0

; delay for A tics...
timeout:
	push hl
	ld hl,tic_ctr
	add a,(hl)
timeout1:
	cp (hl)
	jr nz,timeout1	; wait for tic to reach target
	pop hl
	ret

boot_num_err1:
	call con_bell
	jr boot_num_more

; boot by number..
boot_num:
	call con_out	; echo digit
	and 00fh	; convert to bin
	ld d,a		; D = <num>
boot_num_more:
	call con_input
	cp c
	jr z,boot_got_num	; done
	cp '0'
	jr c,boot_num_err1
	cp '9'+1
	jr nc,boot_num_err1
	call con_out
	and 00fh	; ascii to int
	ld b,10
boot_num1:
	add a,d
	jp c,boot_err
	djnz boot_num1	; A += D * 10
	ld d,a
	cp m314_z47_dev_off
	jp nc,boot_err	; must be < 200
	jr boot_num_more

; got number, validate it
boot_got_num:
	ld a,d
	cp m314_z47_device
	jp c,boot_cr	; < 5
	cp m314_z47_device+m314_z47_device_num
	jp nc,boot_cr	; >= 9
	add a,m314_z47_dev_off	; must be 5,6,7,8: add 200?
	ld d,a
	jp boot_cr

mms_boot:
	ld hl,boot_msg
	call msgz_out
	ld sp,boot_stack
	call get_def_dev
	ld c,00dh	; CR
	jr mms_boot_input

not_alpha:	; not digit or alpha, then error
	call con_bell
	; retry input of char
mms_boot_input:
	call con_input	; echo char
	cp c	; term char?
	jp z,boot_cr
	ld e,0
	cp '0'
	jr c,not_digit	; < '0'
	cp '9'+1
	jr c,boot_num	; < '9'+1	; got digit, boot by number...
	; not digit...
not_digit:
	and 05fh	; toupper()
	cp 'Z'+1
	jr nc,not_alpha	; > 'Z', invalid
	cp 'A'
	jr c,not_alpha	; < 'A', invalid
	; got letter, boot by letter...
boot_by_letter:
	call con_out	; echo entry
	call con_out	; why echo twice?
	cp 'B'
	jr c,boot_got_letter	; 'A'... boot default?
	; 'B'..'Z' - do first lookup...
	ld hl,hz_drv_table
	ld b,a		; B = letter
boot_ltr_lookup:
	ld a,(hl)	; possible boot letter
	inc hl
	ld d,(hl)	; prospective device id
	inc hl
	cp b
	jr z,boot_got_letter	; D has device ID...
	or a
	jr nz,boot_ltr_lookup
	; reached end of table...
	; no match found...
	ld d,0
	ld a,(boot_hook)
	cp 076h	; HALT instruction
	jp nz,boot_err
	; try second lookup...
	call boot_ltr_hook
	jp c,boot_err
	; got boot device id in D...
;
; [B]oot [A-Z]-
;              :<string>
;             ' '
;            [0-9]
;            [0-9]:<string>
; [B]oot [0-9]...
;
; D = device id
; on return, E has entered number
boot_got_letter:
	ld a,'-'	; now get unit number
	call con_out
	jr boot_get_unit

; [B]oot [0-9]...
;                :<string>
boot_num_err:
	call con_bell
boot_get_unit:
	call con_input
	cp c
	jr z,boot_cr
	cp ':'
	jr z,boot_colon
	cp ' '
	jr z,boot_sp	; echo sp and keep going...
	cp '0'
	jr c,boot_num_err
	cp '9'+1
	jr nc,boot_num_err
	call con_out
	sub '0'
	ld e,a		; int value of first digit
boot_got_unit:
	call con_input
	cp c
	jr z,boot_cr
	cp ':'
	jr z,boot_colon
	cp ' '
	jr z,boot_sp	; echo sp and keep going...
	; error -
	ld a,007h ; BEL
boot_sp:
	call con_out
	jr boot_got_unit

; got ':' input...
; read string into tmp_buf and return count in B, term char in C
boot_colon:
	ld b,0
	ld hl,tmp_buf
boot_str:
	call con_out	; echo ':' or last char
	call con_input
	inc b
	inc hl
	ld (hl),a
	cp c
	jr nz,boot_str
	xor a
push_str:
	push af	; transfer string to stack! NUL terminated
	inc sp
	dec hl
	ld a,(hl)
	djnz push_str
	; fall through to boot_cr
; done... got string at SP
; D = device base code, E = unit number
boot_cr:
	call con_crlf
boot_wo_cr:
	ld hl,boot_err
	push hl
	call setup_z17
	ld a,d
	add a,e		; device id + unit number...
	ld (boot_dev),a
	ld a,d
	cp z17_device_num
	jr c,z17_boot	; device 0,1,2
	sub z47_device
	cp z47_device_num
	jr c,z47_boot	; device 5,6,7,8
	ld a,(boot_hook)
	cp 076h	; HALT instruction
	jp z,main_boot
	ret

; ----------------------------------
; H17 5.25" Hard-sectored Controller
z17_boot:
	add a,e
	cp z17_device_num
	ret nc
	ld (boot_unit),a
	in a,(gp_port)
	and gp_pri_dev
	ret nz	; no z17 installed...
	ld a,z17_base
	ld (boot_base_port),a
	ld hl,l_h17_sdp	; install our own version of H17_ROM SDP routine
	ld (h17_sdp+1),hl
	ld a,10
	ld b,a
	call set_timeout
	call l_h17_sdp
z17_wait1:
	call h17_whd_d
	call h17_wnh_d
	djnz z17_wait1
	call h17_abort_d
	ld de,boot_loc	; load into RAM boot location
	ld bc,9*256	; 9 sectors to read
	ld hl,0		; sector 0
	call h17_read_d
	ret c
	ld hl,h17_clock_d
	ld (rst08_vec+1),hl
	pop hl
	jp boot_loc

; ---------------------------------------
; H47/Z47 Interface for "Smart" 8" Drives
z47_boot:
	add a,e	; unit number...
	cp z47_device_num
	ret nc
	ld (boot_unit),a	; 0,1,2,3
	rrca
	rrca
	rrca
	inc a	; x00xxxx1b, x01xxxx1b, x10xxxx1b, x11xxxx1b...
	ld e,a	; bit pattern for drive select?
	ld c,z89_pri_z47 ; should be gp_pri_z47?
	call get_pri_base
	ret nz
	ld a,b
	ld (boot_base_port),a
	call set_timeout_5
	ld a,002h	; command?
	call z47_out_p0
	ld a,002h	; z47_wsts_reset?
	call z47_wait_done_out_p1
	ld a,e
	call z47_wait_tr_out_p1
	call z47_wait_tr_in_p1
	and 00ch	; sector size?	; 0,4,8,12
	rrca
	rrca
	inc a	; 1,2,3,4
	ld b,a	; num sectors (half)
	ld hl,boot_loc
	push bc
	call z47_rd_sec	; read first half
	pop bc
	inc e	; z47_wsts_ffset ?
	call z47_rd_sec	; read second half
	call z47_in_p0
	and z47_rsts_errn
	ret nz
	ld (saved_z17p3),a	; A = 00
	jp boot_loc

z47_wait_tr_out_p1:
	ld d,z47_rsts_trn
	jr z47_wait1

z47_wait_done_out_p1:
	ld d,z47_rsts_donen
z47_wait1:
	scf
	push af
; wait for status bit in D... then output A, using port +CY
z47_wait2:
	call z47_in_p0
	and d
	jr z,z47_wait2
	pop af
	jr z47_out_p1

; output B to base_port+0
z47_out_p0:
	or a	; CY = 0
z47_out_p1:
	push bc
	ld b,a	; save A
	ld a,(boot_base_port)
	adc a,0
	ld c,a
	ld a,b	; restore A
	out (c),a
	pop bc
	ret

z47_wait_tr_in_p1:
	call z47_in_p0
	rlca	; CY = z47_rsts_trn
	jr nc,z47_wait_tr_in_p1
	; CY = 1
	jp z47_in_px

; flash ctl bits, then read B*128 bytes into HL...
z47_rd_sec:
	ld a,007h	; set ctl bits...
	call z47_wait_done_out_p1
	xor a		; clear ctl bits...
	call z47_wait_tr_out_p1
	ld a,e		; restore ctl bits?
	call z47_wait_tr_out_p1
z47_rd_sec1:
	ld c,128
z47_rd_sec2:
	call z47_wait_tr_in_p1
	ld (hl),a
	inc hl
	dec c
	jr nz,z47_rd_sec2
	dec b
	jr nz,z47_rd_sec1
	ret

hz_drv_table:
	defb	'B',z17_device
	defb	'C',z37_device	;
	defb	'D',z47_device	;
	defb	'E',z67_device	;
	defb	0

; check for memory... ? find top of memory...
start_rom:
	ld hl,mem_start-256
find_mem_top:
	inc h
	ld a,(hl)
	inc (hl)
	cp (hl)
	ld (hl),a
	jr nz,find_mem_top
	dec hl
	ld sp,hl	; SP = high memory
	push hl		; save mem_top on stack
	ld hl,cmd_loop	; return to this place...
	push hl
	ld c,003h	; baud select... 0003
retry_baud:
	ld a,i5250_lcr_dlab|i5250_lcr_8b	; setup console port (init 8250)
	out (con_lcr_port),a
	xor a		; divisor hi = 00
	out (con_dlm_port),a
	ld a,c		; divisor lo = 03
	out (con_dll_port),a
	rlca		;
	ld c,a		; C *= 2
	ld a,i5250_lcr_8b	; DLAB off
	out (con_lcr_port),a
	xor a
	out (con_ier_port),a	; all intrs off
	in a,(con_dat_port)	; clear any rx condition

	ld hl,crt_init	; clear screen, etc?
	call msgz_out

	; A = 00, a.k.a. 256
	ld b,25	; delay awhile...
rom_start1:
	dec a
	jr nz,rom_start1
	djnz rom_start1

	in a,(con_lsr_port)	; check serial port condition
	rra	; CY = i8250_lsr_rxa
	in a,(con_dat_port)
	rla	; chr | i8250_lsr_rxa - CCCCCCCSb
	sub ('K'<<1)|i8250_lsr_rxa
	jr nz,retry_baud	; failed - retry with 2x slower BAUD...
	inc hl		; a.k.a. crt_init2
	call msgz_out	; reset screen

	; A = 00, from msgz_out - a.k.a. 256
	ld b,15	; delay some more... 15*(256*16+13) ~= 61625 cy, @ 2MHz ~= 30ms
rom_start2:
	dec a		; 4 cy
	jr nz,rom_start2; 12 cy (taken), 7 cy not taken
	djnz rom_start2	; 13 cy (taken), 8 cy end

	xor a
	ld (auto_disable),a
	ld (saved_gp),a
	ld a,0c9h ; RET
	ld (l2004h),a
	ld hl,((h8_bit6|h8_ssen) << 8) | 000h
	ld (saved_z17p3),hl	; saved_z17p3=00, saved_h8fp=050h
	; start clock... then test CPU speed
	rst 8	; take one artificial tic intr
	ld hl,tic_ctr
	ld de,200
	ld a,(hl)
tic_wait1:
	cp (hl)
	jr z,tic_wait1	; wait until next tic
	add a,5		; wait 5 tics... 2.25ms each?
tic_next:
	inc de		; count these loops... 6 cy
	cp (hl)		; 7 cy
	cp (hl)		; 7 cy
	jr nz,tic_next	; 12 cy (taken): delay until 5 tics passed
	ld a,d		; 32 cycles each loop/count, for 5 tics...
	cp 2		; 512..1023,
	jr z,intr_setup
	ld hl,unsupp_cpu	; didn't get expected results... overdiven CPU? slow?
	call msgz_out
intr_setup:
	;                       0ffffh
	; SP -> |     HL     |
	;       +------------+
	;       |     DE     |
	;       |     BC     |
	;       |     AF     |
	;       |     fp     |
	; SP'-> |    reti    | (cmd_loop)
	; save->|            |
	;       |            |
	; fp -> |            |
	;       |            |
	;       +------------+
	;                       00000h
	ex (sp),hl	; return adr in HL, save prev HL
	push de		; save DE
	push bc		; save BC
	push af		; save PSW
	ex de,hl	; get return address in DE
	ld hl,10	; 10 bytes - 5 pushes...
	add hl,sp	; caller's stack frame...
	push hl		; save frame pointer
	push de		; return address into position...
	ld de,saved_h8fp
	ld a,(de)
	cpl
	and h8_2msdis|h8_ssen
	ret z	; called from intr, this is all we do
	ld hl,2
	add hl,sp	; HL = SP+2 ; adjust for pop to discard
	ld (start_sp),hl; restart_sp
	ret		; return to cmd_loop...

crt_init:	defb	80h, 27,'[?2h',27,'Z',0
crt_init2:	defb	27,'z',0

unsupp_cpu:	defb	'Unsupported CPU speed',0

; [S]ubstitue command: S ...
subs_cmd:
	ld hl,subs_msg
	call msgz_out
	ld hl,l2003h	; get dest addr into l2003h
	ld d,00dh	; stop after CR
	call subs_one_hex
	ex de,hl
subs_again:
	call nl_addr_sp_out
	ld a,(hl)
	call hex_out
	call sp_out
subs_retry:
	call get_hex
	jr nc,got_hex
	cp 00dh	; CR
	jr z,subs_next
	cp '-'
	jr z,echo_backup
	cp '.'
	ret z	; done
	call con_bell
	jr subs_retry

subs_next:
	inc hl
	jr subs_again

echo_backup:
	call con_out
	dec hl
	jr subs_again

got_hex:
	ld (hl),000h	; zero out location
got_hex1:
	call con_out
	call convert_hex
	rld
	call get_hex_wait
	jr nc,subs_next	; error or CR
	jr got_hex1

; Wait for CR, or hex digit.
; returns CY = got hex digit
;         ZR = got CR
; does not return until CR or hex digit
get_hex_wait:
	call con_input
	cp 00dh	; CR
	ret z
	call chk_hex
	ccf
	ret c	; digit OK
	call con_bell
	jr get_hex_wait	; try again

; While waiting for input, check autoboot...
prompt_wait:
	call auto_boot
	ld a,(boot_hook)
	cp 076h	; HALT instruction
	call z,auto_boot_hook
con_input2:
	in a,(con_lsr_port)
	rrca		; i8250_lsr_rxa
	jr nc,prompt_wait

con_input:
	in a,(con_lsr_port)
	rrca		; i8250_lsr_rxa
	jr nc,con_input
	in a,(con_dat_port)
	and 07fh
	cp 07fh	; DEL/RUB
	jp z,cmd_loop
	ret

con_bell:
	ld a,007h	; BEL
	; fall through to con_out
con_out:
	push af
con_out1:
	in a,(con_lsr_port)
	and i8250_lsr_txe
	jr z,con_out1
	pop af
	out (con_dat_port),a
	ret

; input a hex number, when done store in (HL-1), if CY: A has first digit
subs_one_hex:
	push hl	; save subs addr
	ld hl,0
another_hex:
	call nc,con_input	; get new digit unless we already have one...
	call chk_hex
	jr c,term_hex	; invalid or term
	call con_out	; echo digit
	call convert_hex
	add hl,hl
	add hl,hl
	add hl,hl
	add hl,hl	; HL *= 16 : make room for new digit
	or l
	ld l,a		;
	jr another_hex
term_hex:
	cp d		; term char?
	jr z,done_hex
	call con_bell	; invalid entry...
	or a
	jr another_hex	; try again

done_hex:
	call con_out	; echo digit
	ex de,hl	; hex number into DE
	pop hl		; get subs addr
	ld (hl),d
	dec hl
	ld (hl),e	; store number - backwards!
	ret

convert_hex:
	sub '9'+1	; -10..-1, 7..12
	jr nc,cvt_hex1	; must have been 'A'..'F'...
	add a,7		; -3..6
cvt_hex1:
	add a,3	; 0..9, 10..15
	ret

; input a hex digit
get_hex:
	call con_input
	; fall through to chk_hex
; CY = invalid hex digit, NC: A = ascii hex digit
chk_hex:
	cp '0'
	ret c	; < '0'
	cp '9'+1
	ccf
	ret nc	; '0'..'9', NC
	and 05fh	; toupper()
	cp 'A'
	ret c	; < 'A'
	cp 'F'+1
	ccf
	ret	; 'A'..'F' or CY if invalid

nl_addr_sp_out:
	call con_crlf	; new line
addr_sp_out:
	ld a,h
	call hex_out
	ld a,l
	call hex_out
sp_out:
	ld a,' '
	jp con_out

hex_out:
	push af
	rlca
	rlca
	rlca
	rlca
	call nibble_out
	pop af
nibble_out:
	and 00fh
	add a,090h
	daa
	adc a,040h
	daa
	jp con_out

con_waitcr:
	call con_input
	cp 00dh
	jr nz,con_waitcr
con_crlf:
	ld a,00dh
	call con_out
	ld a,00ah
	jp con_out

	defs 17,0

l0613h:
	jp z47_wait_tr_out_p1

	defb 0

l0617h:
	jp z47_wait_done_out_p1

msgz_out:
	ld a,(hl)
	or a
	ret z
	call con_out
	inc hl
	jr msgz_out

bad_cksum:
	ld hl,bad_cksum_msg
	jp msgz_out

bad_cksum_msg:
	defb	7,'Checksum error',0
top_mem_msg:
	defb	'Top of Memory: ',0

mms_memtest:
	ld hl,memtest_msg
	call msgz_out
	call con_waitcr
	ld hl,top_mem_msg
	call msgz_out
	ld hl,0
	add hl,sp
	ld a,h
	inc a
	jr z,got_64k	; H was FF - 64K already banked-in
	sub 8192>>8	; else subtract 8K - for ROM image???
got_64k:
	ld h,a
	ld l,0
	dec hl		; HL = last byte to test?
	sub memtest_loc>>8
	ld e,a		; E = page to flip-flop around: -12K from top
	call addr_sp_out
	call con_crlf
	ld d,0	; initialize memtest params...
	ld c,memtest_loc>>8
	ld b,0
	; C = ?
	; B = ?
	; E = dividing page
	; D = ?
	; HL = top?
	exx		; memtest will pick up params from alt regs
	ld hl,memtest_seq
	ld de,memtest_loc-(memtest_seq_cksum-memtest_seq)
	ld bc,memtest_seq_len
	ldir
	; checksum memtest before jumping into oblivion...
	ld de,memtest_loc
	ld hl,memtest_seq_cksum
	ld c,memtest_seq_cksum_len
	xor a
	ex af,af'
	xor a
memtest_cksum:
	add a,(hl)
	ex af,af'
	ex de,hl
	add a,(hl)
	ex af,af'
	ex de,hl
	inc hl
	inc de
	dec c
	jp nz,memtest_cksum
	ld c,a
	ex af,af'
	cp c
	jp nz,bad_cksum
	di
	jp memtest_loc-(memtest_seq_cksum-memtest_seq_start)

memtest_seq:
	; start memory test, position-independent code copied into RAM...

mms_magic_seq:	; this unlocks the low 8K of RAM (swaps out ROM) on MMS 77312
	defb	004h,00ch,004h,008h,00ch,008h,020h
mms_magic_len:	equ	$-mms_magic_seq

memtest_seq_start:
	ld hl,memtest_loc-(memtest_seq_cksum-memtest_seq)
	ld bc,(mms_magic_len<<8)|gp_port
	otir

memtest_seq_cksum:
; This gets loaded int 03000h, and alternately into 00000h
mtest_start:
	exx
	ld h,d
	ld l,0
	ld a,b
	exx
	ld c,a
	ld b,2
; print B hex digits of C..., then CR
mtest_hexout1:
	ld a,c
	rlca
	rlca
	rlca
	rlca	; non-destructive flip-flop nibbles
	ld c,a
	and 00fh
	add a,090h
	daa
	adc a,040h
	daa
	out (con_dat_port),a
mtest_conout1:
	in a,(con_lsr_port)
	and i8250_lsr_txe
	jr z,mtest_conout1
	dec b
	jr nz,mtest_hexout1
	ld a,00dh	; CR
	out (con_dat_port),a
	exx
	; C = num pages to test
	ld a,b	; B = starting seq number
	; memtest pattern is BCD sequence, this avoids patterns that might
	; hide errors related to address lines, etc.
mtest_fill_pg:
	ld (hl),a
	add a,1
	daa
	inc l
	jr nz,mtest_fill_pg
	inc h
	dec c
	jr nz,mtest_fill_pg
	ld a,h
	sub d	; recompute num pages
	ld c,a
	ld h,d	; reset address
	ld l,0
	ld a,b
mtest_chk_pg:
	cp (hl)
	jr nz,mtest_mem_err
	add a,1
	daa
	inc l
	jr nz,mtest_chk_pg
	inc h
	dec c
	jr nz,mtest_chk_pg
	exx
	ld hl,memtest_loc
	ld de,0
	ld bc,memtest_seq_cksum_len
	exx
	ld a,d
	xor memtest_loc>>8	; flip-flop between there and 00000h
	ld d,a
	jr z,mtest_bank0
	; testing bank 1...
	ld c,e
	jr mtest_self_reloc
mtest_bank0:
	; reset and change starting seq value
	ld c,memtest_loc>>8 ; too simple?
	ld a,1
	add a,b
	daa
	ld b,a	; new starting seq value
	exx
	ex de,hl	; flip address/length
	exx
mtest_self_reloc:
	exx
	ldir	; copy code to alternate bank
	ld a,d
	and 0f0h
	ld h,a
	ld l,0
	ld c,memtest_seq_cksum_len
	xor a
mtest_chk_sum:
	add a,(hl)
	inc hl
	dec c
	jr nz,mtest_chk_sum
	ld c,a
	ex af,af'
	cp c
	jr nz,mtest_chk_fail	; checksum failed
	ex af,af'
	ld a,d
	and 0f0h
	ld h,a
	ld l,0
	jp (hl)	; jump to newly-relocated mtest code

mtest_mem_err:
	xor (hl)	; get bits in error
	ld d,a
	ld a,00ah	; LF
	out (con_dat_port),a
mtest_conout6:
	in a,(con_lsr_port)
	and i8250_lsr_txe
	jr z,mtest_conout6
	ld c,002h
	ld b,4
mtest_hexout2:
	ld a,h
	rlca
	rlca
	rlca
	rlca
	and 00fh
	add a,090h
	daa
	adc a,040h
	daa
	out (con_dat_port),a
mtest_conout4:
	in a,(con_lsr_port)
	and i8250_lsr_txe
	jr z,mtest_conout4
	add hl,hl
	add hl,hl
	add hl,hl
	add hl,hl
	djnz mtest_hexout2
	ld a,' '
	out (con_dat_port),a
mtest_conout5:
	in a,(con_lsr_port)
	and i8250_lsr_txe
	jr z,mtest_conout5
	dec c
	ex de,hl
	ld b,2
	jr nz,mtest_hexout2
	ld a,'*'
	out (con_dat_port),a
	jr mtest_die

mtest_chk_fail:
mtest_conout2:
	in a,(con_lsr_port)
	and i8250_lsr_txe
	jr z,mtest_conout2
	ld a,00ah ; LF
	out (con_dat_port),a
mtest_conout3:
	in a,(con_lsr_port)
	and i8250_lsr_txe
	jr z,mtest_conout3
	ld a,'!'
	out (con_dat_port),a
mtest_die:
mtest_conout7:
	in a,(con_lsr_port)
	and i8250_lsr_txe
	jr z,mtest_conout7
	xor a
	ld b,250	; delay 64000
mtest_delay:
	dec a
	jr nz,mtest_delay
	djnz mtest_delay
	ld a,007h	; BEL
	out (con_dat_port),a
	jr mtest_die
memtest_seq_len:	equ	$-memtest_seq
memtest_seq_cksum_len:	equ	$-memtest_seq_cksum

setup_z17:
	di
	push de
	ld hl,saved_h8fp
	ld (hl),h8_bit7|h8_bit6|h8_ssen	; !h8_2msdis
	ld hl,h17_block_src
	ld de,h17_block_dest
	ld bc,h17_block_len
	ldir
	ld l,e	; DE = h17_zero_init
	ld h,d
	inc de
	ld c,30
	ld (hl),0
	ldir	; bzero(HL,31)
	xor a		; A = 00
	ld (boot_unit),a
	out (z17_port3),a
	inc a		; A = 01
	ld (saved_z17p3),a
	ld hl,rst_vecs
z17_set_jmps:
	ld (hl),0c3h	; JMP ei_ret
	inc hl
	ld (hl),ei_ret & 000ffh
	inc hl
	ld (hl),ei_ret >> 8
	inc hl
	add a,a		; A <<= 1
	jp p,z17_set_jmps	; 7 times...
	pop de
	ret

; BLOCK 'data40' (start 0x07d4 end 0x07fa)
mms_prompt:
	defb	0dh,0ah,'MMS: ',0
	defb	'(c) 1982 MMS',0

	defb	0
	defs	16,0

; BLOCK 'text50' (start 0x07fa end 0x0b46)
get_pri_hook:
	jp get_pri_base
set_timeout_hook:
	jp set_timeout
boot_hook:
	halt
auto_boot_hook:
	jp dummy_fn1
rom_cmd_hook:
	jp dummy_fn2
main_boot:
	jp boot
def_dev_hook:
	jp mms_def_dev
boot_ltr_hook:
	jp search_mms_drv
dummy_fn1:
	ret
dummy_fn2:
	ret

; D = device id number, E = unit
boot:
	ld a,d
	sub m314_z47_dev_off	; convert m314_z47 device
	jr c,disk_boot
	ld d,a
disk_boot:
	; lookup device number in table, get boot routine.
	ld hl,boot_table
find_device:
	ld a,d
	sub (hl)
	inc hl
	cp (hl)
	jr c,found_device
	ld a,(hl)
	inc hl
	inc hl
	inc hl
	or a
	jr nz,find_device
	ret	; ZR = device not found
found_device:
	inc hl
	ld c,(hl)
	inc hl
	ld h,(hl)
	ld l,c
	add a,e
	jp (hl)	; go to specific device boot routine

; port F2H: xAAAxxxx
; examine config for boot device hint...
; returns ZR if none installed.
mms_def_dev:
	in a,(gp_port)	; get config switches...
	and gp_def_boot
	rlca
	rlca
	rlca
	rlca
	ld hl,gp_table
; get boot device for index A
mms_get_dev:
	add a,l		; get &HL[A]
	ld l,a
	ld a,0
	adc a,h
	ld h,a
	ld a,(hl)	; get HL[A]
	cp invalid_device
	ret z
	cp sub_device
	jp z,try_m314_gp
	ld d,a
	ret

; returns E = unit number, D = device id (base) number
try_m314_gp:
	in a,(m314_gp)	; more config?
	and m314_gp_boot
	rlca
	rlca
	rlca
	ld hl,m314_gp_table
	call mms_get_dev	; recurse on sub-device table
	ret z	; none found
	in a,(m314_gp)
	and m314_gp_opts	; specific 77316 unit number
	rrca
	rrca
	ld e,a
	ret

; search for match, B = target device letter
; returns D = base device id, CY = not found
search_mms_drv:
	ld hl,mms_drv_table
mms_drv_loop:
	ld a,(hl)
	inc hl
	ld d,(hl)
	inc hl
	cp b
	ret z
	or a
	jr nz,mms_drv_loop
	ld d,0
	scf
	ret

; z17 and z47 (heath/zenith) devices already checked
boot_table:
	;	z17_device,z17_device_num
	;		z17_boot
	defb	z67_device,z67_device_num
	defw		z67_boot
	defb	m314_z47_device,m314_z47_device_num
	defw		m314_z47_boot
	defb	m314_imi_device,m314_imi_device_num
	defw		m314_imi_boot
	defb	m316_device,m316_device_num
	defw		m316_boot
	defb	x37_device,x37_device_num
	defw		x37_boot
	defb	x38_device,x38_device_num
	defw		x38_boot
	defb	z37_device,z37_device_num
	defw		z37_boot
	defb	0,0	; END of table

gp_table:	; gp_xxx654 indexes into this...
	defb	m316_5_device	; gp_boot_mms5 - use first MMS 5" floppy
	defb	m316_8_device	; gp_boot_mms8 - use first MMS 8" floppy
	defb	invalid_device	; gp_boot_pri - already checked
	defb	invalid_device	; gp_boot_sec - already checked
	defb	invalid_device	; gp_xxx654_4
	defb	invalid_device	; gp_xxx654_5
	defb	invalid_device	; gp_xxx654_6
	defb	sub_device	; gp_boot_hdd	; use m314_gp

m314_gp_table:	; (m314_gp) 11100000b indexes this...
	defb	m314_imi_device	; m314_gp_imi
	defb	m314_z47_device+m314_z47_dev_off	; m314_gp_1
	defb	x38_device	; m314_gp_2
	defb	x37_device	; m314_gp_3
	defb	invalid_device	; m314_gp_res4
	defb	invalid_device	; m314_gp_res5
	defb	invalid_device	; m314_gp_res6
	defb	invalid_device	; m314_gp_res7

mms_drv_table:
	defb	'G',m314_z47_device+m314_z47_dev_off
	defb	'H',m314_imi_device
	defb	'I',m316_8_device
	defb	'J',m316_5_device
	defb	'K',x37_device
	defb	'L',x38_device
	defb	0

; --------------------------------
; MMS 77314 IMI Harddisk Interface
m314_imi_boot:
	ld d,a	; unit num
	ld a,10
	call set_timeout_hook
	in a,(m314_imi_status)
	and m314_imi_power
	ret nz
	ld b,000h ; a.k.a. 256
	in a,(m314_imi_status)
	ld c,a
	ld a,0ffh	; no-op command?
	out (m314_imi_data),a
imi_wait_sts:
	in a,(m314_imi_status)
	cp c
	jr nz,imi_sts_chg
	djnz imi_wait_sts	; wait for "something" to change...
	ret	; failed

imi_sts_chg:
	ld b,000h
imi_retry:
	ex (sp),hl	;
	ex (sp),hl	; just waste some time?
	ld a,0ffh	; no-op command?
	out (m314_imi_data),a
imi_wait_rdy:
	in a,(m314_imi_status)
	rrca	; m314_imi_ready
	jr c,imi_wait_rdy
	rrca	; m314_imi_ifact
	jr c,imi_if_active
	djnz imi_retry
	ret	; failed

imi_if_active:
	call imi_recv
	cp 08fh		; command response - drive type?
	ret nz
	ld iy,boot_loc
	ld a,d
	cp m314_imi_device_num
	ret nc	; invalid unit number
	; partition table at sector 0, simple array of 3-byte integers
	; of starting sector address for each virtual drive.
	ld d,a
	add a,a	;
	add a,d	; a = d * 3 : len of each partition table entry
	ld c,a
	ld b,0
	add iy,bc	; index into (not-yet-read) partition table
	ld ix,boot_loc
	ld bc,0100h	; B=1 (num sectors), C=0;
	ld de,0		; sector address 0
	call imi_read	; get partition table
	ret c
	ld a,(iy+0)
	rla
	ret c	; noy bootable? not present?
	ld c,(iy+0)	;
	ld d,(iy+1)	;
	ld e,(iy+2)	; get sector address of partition
	ld ix,boot_loc
	ld b,2		; num sectors
	call imi_read	; now boot from partition
	ret c
	jp boot_loc

; read some sectors from the IMI drive
; B = num sectors, C = sec19-16, D = sec15-8, E = sec7-0
imi_read:
	ld a,012h	; read command
	call imi_send
	ld a,c		; hi 4 bits of sector address
	add a,a
	add a,a
	add a,a
	add a,a
	inc a		; C * 16 + 1 = CCCC0001b
	call imi_send
	ld a,e		; more sector address
	call imi_send
	ld a,d		; last part of sector address
	call imi_send
imi_wait_act:
	in a,(m314_imi_status)
	and m314_imi_ifact
	jr z,imi_wait_act
	ld a,8
imi_delay:
	dec a
	jp nz,imi_delay	; a short delay
	call imi_recv
	rlca	; CY = m314_imi_power
	ret c		; error: lost power
	ld l,128	; byte count - sector size
imi_sec_loop:
	call imi_recv
	ld (ix+000h),a
	inc ix
	dec l
	jr nz,imi_sec_loop
	inc e	; increment sector address
	jr nz,imi_sec_nocy
	inc d	; carry over
	jr nz,imi_sec_nocy
	inc c	; carry over
imi_sec_nocy:
	djnz imi_read	; count sector and read another
	or a
	ret

; Send one byte of data or command to IMI drive, wait for READY
imi_send:
	push af
imi_rdy_loop:
	in a,(m314_imi_status)
	rra	; CY = m314_imi_ready = Not Ready
	jr c,imi_rdy_loop
	pop af
	out (m314_imi_data),a
	ret

; Get one byte of data or results from IMI drive, wait for READY
imi_recv:
	in a,(m314_imi_status)
	rra	; CY = m314_imi_ready = Not Ready
	jr c,imi_recv
	in a,(m314_imi_data)
	ret

; ------------------------------------
; MMS 77316 5.25"/8" Floppy Controller
;
; completion intr jams "RST 30" onto bus,
; data req intr jams "EI" onto bus - least overhead, simply un-HALTs us.
; NOTE! we must be in a sub-routine when the completion intr happens,
; since it discards the intr return and effectively causes us to return.
m316_boot:
	ld hl,m316_intr	; completion interrupt, install at RST 30
	ld (rst30_vec+1),hl
	cp m316_device_num
	ret nc
	or m316_gp_ie|m316_gp_burstn
	ld d,a
	out (m316_gp),a
	xor a
	out (gp_port),a
	ld a,wd1797_cmd_restore|wd1797_cmd_load|wd1797_cmd_step3
	call m316_start_cmd
	ld bc,53000
wait_ready:
	in a,(m316_cmd)
	rlca	; wd1797_sts1_nrdy
	jr nc,disk_ready
	dec bc
	ld a,b
	or c
	jr nz,wait_ready
disk_ready:
	in a,(m316_cmd)
	and wd1797_sts1_nrdy|wd1797_sts1_sker|wd1797_sts1_crer|wd1797_sts1_bsy
	ret nz
	ld e,25	; retry count
retry_boot:
	ld hl,boot_loc
	xor a
get_more:
	inc a
	out (m316_sec),a
	ld bc,m316_data
	ld a,wd1797_cmd_read|8	; "8" is obsolete/ignored bit in command
	call m316_read
	and wd1797_sts2_nrdy|wd1797_sts2_wflt|wd1797_sts2_rnf|wd1797_sts2_crer|wd1797_sts2_lost|wd1797_sts2_dreq|wd1797_sts2_bsy
	ld a,d
	out (m316_gp),a
	jp z,m316_chk_sec	; success...
	xor m316_gp_ddn	; try single density, just for kicks
	ld d,a
	dec e
	jr nz,retry_boot
	ret	; failed.

m316_chk_sec:
	in a,(m316_sec)
	cp 2	; read 2 sectors, sometimes overkill...
	jr c,get_more
	jp boot_loc

m316_read:
	push af
	ld a,d
	and m316_gp_ddn | m316_gp_5in
	jr nz,m316_read_noburst	; 5" drives or single density doesn't use burst mode
	ld a,d
	and ~m316_gp_burstn	; set BURST mode?
	out (m316_gp),a
	pop af
	out (m316_cmd),a	; issue command to WD1797
	ei
	halt	; wait for first byte to arrive...
m316_read_loop:	; we break out of loop via completion interrupt
	ini	; WAIT states keep us in-sync
	jp m316_read_loop

m316_read_noburst:
	pop af
	out (m316_cmd),a	; issue command to WD1797
z37_bulk_read:	; convenient re-use of code
	ei
m316_z37_loop:	; we break out of loop via completion interrupt
	halt	; "NIL" interrupt signals byte ready
	ini
	jp m316_z37_loop

m316_start_cmd:
	out (m316_cmd),a
z37_wait_intr:
	ei
m316_z37_stall:	; we break out of loop via completion interrupt
	jr m316_z37_stall

m316_intr:
	pop af	; discard return address - pop out of loop
	in a,(m316_cmd)
	ei
	ret

; -------------------------------
; H37/Z37 5.25" Floppy Controller
z37_boot:
	; install intreq vector for z37
	ld hl,z37_intr
	ld (rst20_vec+1),hl
	dec hl		; HL = z37_dummy
	ld (z37_ret),hl
	cp z37_device_num
	ret nc
	ld (boot_unit),a
	inc a		; drive + 1: {1..4}
	; compute drive-select bits
	ld l,z37_icl_ds0>>1	; drive select
z37_drvsel:
	add hl,hl
	dec a
	jr nz,z37_drvsel

	out (z37_acl),a		; A = 00: known state
	in a,(gp_port)
	and gp_sec_dev
	ret nz		; no z37 installed
	ld a,z37_base
	ld (boot_base_port),a
	ld a,wd1797_cmd_intr
	out (z37_cmd),a
	ld a,l		; drive select bits
	or z37_icl_motor
	ld d,a
	out (z37_icl),a	; drive select to data port??
	inc d		; D |= z37_icl_inten
	ld e,019h
	ld a,5
	call set_timeout_hook

	; Wait for index hole to pass by a couple times...
	; it must be changing state.
	ld bc,(20 << 8)|z37_data
z37_wait_idx:
	in a,(z37_cmd)
	xor b
	and wd1797_sts1_indx
	jr z,z37_wait_idx
	djnz z37_wait_idx

z37_retry:
	ld hl,boot_loc
	ld a,001h
	out (z37_acl),a	; select trk/sec regs
	out (z37_sec),a	; set sector to 1
	ld a,d
	out (z37_icl),a
	xor a
	out (z37_acl),a	; select cmd/data regs
	ld a,wd1797_cmd_restore|wd1797_cmd_load|wd1797_cmd_step3
	out (z37_cmd),a
	call z37_wait_intr
	ld a,d
	xor z37_icl_dden
	ld d,a
	or z37_icl_drqen
	out (z37_icl),a
	ld a,wd1797_cmd_read|wd1797_cmd_mult|wd1797_cmd_slen|wd1797_cmd_15ms
	out (z37_cmd),a
	call z37_bulk_read	; read entire track
	and wd1797_sts2_nrdy|wd1797_sts2_prot|wd1797_sts2_wflt|wd1797_sts2_crer|wd1797_sts2_lost|wd1797_sts2_dreq|wd1797_sts2_bsy
	jr nz,z37_bad_sts
	ld a,h
	cp 02ch	; 02280h+0a00h ; (boot_loc+0a00h)>>8
	jr c,z37_bad_sts	; did not load enough...
	ld a,z37_icl_motor	; leave motor running, everything else off...
	out (z37_icl),a
	ld hl,h17_clock_d	; restore vector
	ld (rst08_vec+1),hl
	pop hl
	jp boot_loc

z37_bad_sts:
	dec e		; error count
	jr nz,z37_retry	; retry
	; return failure...
z37_dummy:
	ret

z37_intr:
	in a,(z37_cmd)
	ex (sp),hl
	ld hl,(z37_ret)
	ex (sp),hl
	ei
	ret

; --------------------
; Unknown device id 37
x37_boot:
	or a
	ret nz
	out (x37_port4),a	; 00
	out (x37_port0),a	; 00
	out (x37_port1),a	; 00
	call x37_runout	; clear out buffers?
	ld hl,x37_init_seq
	; process I/O data "program"...
	ld a,(hl)
	out (x37_port4),a
	inc hl
	ld a,(hl)
	out (x37_port5),a
	inc hl
	ld c,x37_port6
x37_init_loop:
	ld a,(hl)
	or a
	jr z,x37_init_done
	inc hl
	ld b,a		; count and/or repeat flag
	rla
	jr c,x37_init_rpt	; do repeat
	otir		; else output B bytes
	jr x37_init_loop

x37_init_rpt:
	or a	; clear CY
	rra	; get repeat count w/o flag
	ld b,a
	ld a,(hl)
	inc hl
x37_init_rpt1:
	out (x37_port6),a
	djnz x37_init_rpt1
	jr x37_init_loop

; done with I/O sequence,
x37_init_done:
	ld a,041h
	out (x37_port1),a	; 41
	ld a,002h
	out (x37_port0),a	; 02
	call x37_wait_bit2	; while sending  00 01 03
	ret nz	; timeout
	call x37_read_cmd
	ret nz	; timeout
	; data is waiting, no more errors possible...
	ld a,002h
	out (x37_port4),a	; 02
	xor a
	out (x37_port5),a	; 00
	ld b,a	; 00 a.k.a. 256
	ld c,x37_port6
	ld hl,boot_loc
	in a,(x37_port6)	; discard first byte...
	inir
	jp boot_loc

x37_read_cmd:
	ld d,10		; 10 attempts
	ld a,004h
	out (x37_port4),a	; 04
	ld a,0eah
	out (x37_port5),a	; EA
	ld b,4		; spit out 4 "00"
	xor a
x37_cmd1:
	out (x37_port6),a	; 00
	djnz x37_cmd1
x37_retry:
	call x37_test_bit0
	ret nz
	xor a
	out (x37_port4),a	; 00
	ld a,0d7h
	out (x37_port5),a	; D7
	ld a,008h
	out (x37_port4),a	; 08
x37_cmd2:
	in a,(x37_port4)
	rrca
	jp nc,x37_cmd2	; wait for bit0 == 1
	nop
	in a,(x37_port4)
	ld b,a
	xor a
	out (x37_port4),a	; 00 (was 08)
	ld a,b
	and 00eh
	ld b,a
	in a,(x37_port0)
	and 002h
	call nz,x37_pt1_00_41
	or b	; can only be zero if call not taken AND B==0
	ret z
	dec d
	jr nz,x37_retry
x37_ret_1:
	or 001h
	ret

; ZR = success, NZ/01 = timeout
x37_wait_bit2:
	ld b,000h
x37_wait1:
	in a,(x37_port0)
	and 004h
	xor 004h
	ret z		; bit2 == 1
	dec b
	jr z,x37_ret_1	; return 1 - timeout
	xor a
	out (x37_port3),a	; 00
	inc a
	out (x37_port2),a	; 01
	ld a,003h
	out (x37_port0),a	; 03
	call x37_wait_bit7		; wait for !busy?
	jp x37_wait1

; wait for (bit7 == 1)
x37_wait_bit7:
	in a,(x37_port0)
	rla
	jr nc,x37_wait_bit7
	ret

; ZR = (bit0 == 1)
x37_test_bit0:
	in a,(x37_port0)
	and 001h
	xor 001h
	ret

; send 00,41 to x37_port1
x37_pt1_00_41:
	xor a
	out (x37_port1),a	; 00
	ld a,041h
	out (x37_port1),a	; 41
	ret

; output 00 to x37_port4, 128 "0F" to port x37_port6,
; then   01 to x37_port4, 128 "0F" to port x37_port6
x37_runout:
	xor a
	call x37_runout1
	ld a,001h
x37_runout1:
	out (x37_port4),a
	ld b,128
	ld a,00fh
x37_runout2:
	out (x37_port6),a
	djnz x37_runout2
	ret

; BLOCK 'data50' (start 0x0b46 end 0x0b70)
x37_init_seq:
	defb	000h	; init port x37_port4
	defb	0d7h	; init port x37_port5
	; rest are sent to port x37_port6...
	;   0-127: output the next N bytes to port x37_port6
	; 128-255: output the next byte N-128 times to port x37_port6
	defb	002h,040h,041h	; next 2 bytes
	defb	083h,040h	; repeated 3 times...
	defb	08ch,06dh	; repeat 12...
	defb	001h,063h
	defb	084h,065h
	defb	082h,067h
	defb	001h,049h
	defb	082h,040h
	defb	08ch,06dh
	defb	003h,063h,0e7h,067h
	defb	087h,00fh
	defb	002h,040h,041h
	defb	083h,040h
	defb	08ch,06dh
	defb	001h,063h
	defb	082h,065h
	defb	082h,060h
	defb	082h,067h
	defb	001h,00fh
	defb	0 ; END

; --------------------
; Unknown device id 38
x38_boot:
	cp x38_device_num
	ret nc
	out (x38_port5),a	; unit select
	xor a
	call x38_send_cmd	; cmd 0 - reset?
	ld a,006h
	call x38_send_cmd	; cmd 6 - seek 0?
	xor 064h
	and 066h
	ret nz
	out (x38_port1),a	; A = 00
	out (x38_port2),a	;
	out (x38_port3),a	; sector?
	ld a,001h
	call x38_send_cmd	; cmd 1 - read?
	ld hl,boot_loc
	ld bc,x38_port4	; B = 0, C = x38_port4
	in a,(x38_port4)
	inir		; 256 bytes from port x38_port4...
	in a,(x38_port0)
	xor 055h
	and 0dfh
	ret nz
	jp boot_loc

; send cmd, wait for !busy, return status
x38_send_cmd:
	out (x38_port0),a	; send cmd
x38_wait_nbsy:
	in a,(x38_port0)
	rla
	jr c,x38_wait_nbsy
	in a,(x38_port0)
	ret

; ---------------------------------
; H67/Z67 Winchester Disk Interface
z67_boot:
	cp z67_device_num
	ret nc	; >= 2 invalid drive for z67...
	ld (boot_unit),a	; 0,1
	rrca
	rrca
	rrca
	ld d,a	; unit select... 00U00000b
	ld c,z89_pri_z67
	call get_pri_hook
	ret nz		; no H67 installed...
	ld a,b
	ld (boot_base_port),a
	inc a		; z67_base + 1
	ld c,a
	xor a
	out (c),a
	ld hl,0
	ld (z67_cmd_buf),hl
	ld (z67_cmd_buf+2),hl
	ld (z67_cmd_buf+4),hl
	ld a,4		; initial delay - short
	or a
	ei
z67_retry:
	ret z
	call delay
	ld e,000h	; cmd 0 = ??? reset?
	call z67_send_cmd
	ld a,255	; retry delay, long
	jr c,z67_retry

	ld e,001h	; cmd 1 = seek???
	call z67_send_cmd
	ret c
	ld hl,0800ah	; 128 || 10
	ld (z67_cmd_buf+4),hl
	ld e,008h	; cmd 8 = read
	call z67_send_cmd
	ret c
	pop hl
	jp boot_loc

; C = I/O port base, E = cmd byte, D = param byte (unit)
; CY = timeout, ZR = success
z67_send_cmd:
	di
	ld l,e
	ld h,d
	ld (z67_cmd_buf),hl
	ld b,0		; !z67_sts_bsy
	ld e,6		; wait...
	ld hl,0		; count = 6 * 65536
; set SEL and wait for !busy -> busy
z67_wait1:
	in a,(c)
	and z67_sts_bsy
	cp b
	jr z,z67_ready
	dec hl
	ld a,l
	or h
	jr nz,z67_wait1
	dec e
	jr nz,z67_wait1
	scf
	ret		; failed to see !busy
z67_ready:
	ld a,b
	xor z67_sts_bsy	; now look for !busy
	jr z,z67_got_busy	; done: saw full transition
	ld b,a
	ld a,z67_ctl_sel
	out (c),a
	jr z67_wait1	; keep waiting...

; saw !busy -> busy
z67_got_busy:
	ld a,z67_ctl_daten	; (and !z67_ctl_sel)
	out (c),a
	ld hl,z67_cmd_buf	; command buffer?
z67_wait2:
	in a,(c)
	bit 7,a		; z67_sts_req
	jr z,z67_wait2
	bit 4,a		; z67_sts_cmd
	jr z,z67_get_data
	bit 6,a		; z67_sts_inout
	jr z,z67_done
	dec c
	outi	; output command byte
	inc c
	jr z67_wait2

; command sent, now get data back...
z67_get_data:
	ld hl,boot_loc
z67_wait3:
	in a,(c)
	bit 7,a		; z67_sts_req
	jr z,z67_wait3
	bit 4,a		; z67_sts_cmd
	jr nz,z67_done
	dec c
	ini	; get data byte
	inc c
	jr z67_wait3

z67_done:
	in a,(c)
	and z67_sts_req|z67_sts_inout|z67_sts_cmd
	cp z67_sts_req|z67_sts_cmd
	jr nz,z67_done
	dec c
	in l,(c)	; result bytes
	inc c
z67_wait4:
	in h,(c)
	ld a,h
	and z67_sts_req|z67_sts_inout|z67_sts_msg
	cp z67_sts_req|z67_sts_msg
	jr nz,z67_wait4
	ld (z67_res_buf),hl
	dec c
	in a,(c)	; final byte???
	inc c
	ei
	or a
	scf
	ret nz		; error?
	bit 0,l	; parity error?
	ret nz
	bit 1,l	; xxx error?
	ret nz
	bit 1,h	; z67_sts_irq
	ret nz
	xor a
	ret

; --------------------------------------------------
; MMS 77314 Remex/Z47 Interface for "Smart" 8" Disks
m314_z47_boot:
	cp m314_z47_device_num
	ret nc
	rrca
	rrca
	rrca	; move unit select into position
	inc a	; add in sector number (1)
	ld d,a
	ld a,5
	call set_timeout_hook
	in a,(m314_z47_status)
	and m314_z47_power
	ret nz
	out (m314_z47_strobe),a
m314_z47_wait1:
	in a,(m314_z47_status)
	rla	; m314_z47_bsy
	jr c,m314_z47_wait1
m314_z47_more:
	ld a,007h	; read command...
	out (m314_z47_data),a	; command 07
	xor a
	call m314_z47_send_cmd	; 00 - track?
	ld a,d	; sector number with unit sel
	call m314_z47_send_cmd	; UUU0SSSSb, unit + sector
m314_z47_wait2:
	in a,(m314_z47_status)
	and m314_z47_bsy
	jr z,m314_z47_wait2
m314_z47_wait3:
	in a,(m314_z47_status)
	and m314_z47_dtr
	jr nz,m314_z47_wait3
	in a,(m314_z47_status)
	and m314_z47_err
	ret nz
	ld hl,boot_loc
	ld b,128
	ld c,m314_z47_data
m314_z47_sec2:
	in a,(m314_z47_status)
	and m314_z47_dtr
	jr nz,m314_z47_sec2
	ini
	jr nz,m314_z47_sec2
	ld b,0	; a.k.a. 256
m314_z47_wait4:
	in a,(m314_z47_status)
	and m314_z47_dtr
	jr z,m314_z47_read_lgsec	; data still coming, read more
	djnz m314_z47_wait4
	in a,(m314_z47_status)
	and m314_z47_err
	ret nz
	inc d
	ld a,d
	and 00fh
	cp 3	; 2 sectors to load
	jr c,m314_z47_more
	jp boot_loc

m314_z47_read_lgsec:
	ld b,128
m314_z47_sec1:
	in a,(m314_z47_status)
	and m314_z47_dtr
	jr nz,m314_z47_sec1
	ini
	jr nz,m314_z47_sec1
	in a,(m314_z47_status)
	and m314_z47_err
	ret nz
	jp boot_loc

m314_z47_send_cmd:
	push af
m314_z47_cmd1:
	in a,(m314_z47_status)
	and m314_z47_dtr|m314_z47_ddout
	jr nz,m314_z47_cmd1
	pop af
	out (m314_z47_data),a
	ret

	defs 766,0ffh
	defw 0,0

	end
