if ?__z89_inc__ == 0
__z89_inc__: equ 1

	include 'wd1797.inc'
	include 'i8250.inc'

; Z89/H89 platform defines

cass_base:	equ	0f8h
con_base:	equ	0e8h
prt_base:	equ	0e0h
mdm_base:	equ	0d8h
alt_base:	equ	0d0h

z89_pri_base:	equ	07ch
z89_sec_base:	equ	078h

z17_base:	equ	z89_pri_base
z47_base2:	equ	z89_pri_base
z47_base1:	equ	z89_sec_base
z37_base:	equ	z89_sec_base
z67_base2:	equ	z89_sec_base
z67_base1:	equ	z89_pri_base

; H8/Legacy support
h8_fp0:		equ	0f0h	; H8 front panel
h8_fp1:		equ	0f1h	; H8 front panel
h8_con0:	equ	0fah	; H8 console
h8_con1:	equ	0fbh	; H8 console

h8_bit7:	equ	10000000b
h8_bit6:	equ	01000000b
h8_2msdis:	equ	00100000b	; 2ms "tic" interrupt disable
h8_ssen:	equ	00010000b	; single-step enable

gp_port:	equ	0f2h	; H89 gp port

; MMS devices...
m316_gp:	equ	038h	; MMS 77316 Floppy gp port
m316_base:	equ	03ch	; MMS 77316 Floppy Controller

x37_base:	equ	048h	; MMS device?
x38_base:	equ	050h	; MMS device?

m314_gp:	equ	05ch	; MMS 77314 IMI/REMEX(Z47) gp port
m314_base:	equ	058h	; MMS 77314 IMI/REMEX(Z47) interface

m320_base:	equ	000h	; MMS 77320 SASI Interface

; H89 gp port --------------------------------------
; H89 gp port dipswitches (input)

gp_autoboot:	equ	10000000b
gp_autoboot_b:	equ	7

; these were usurped to support more device installation config...
gp_def_boot:	equ	01110000b	; MMS: default boot device
gp_boot_hdd:	equ	01110000b	; boot from a MMS 77314 device? '320 too?
gp_xxx654_6:	equ	01100000b
gp_xxx654_5:	equ	01010000b
gp_xxx654_4:	equ	01000000b
gp_boot_sec:	equ	00110000b	; whatever sec port device there is...
gp_boot_pri:	equ	00100000b	; whatever pri port device there is...
gp_boot_mms8:	equ	00010000b	; MMS 77316 8" floppy
gp_boot_mms5:	equ	00000000b	; MMS 77316 5" floppy

gp_con_fast:	equ	01000000b	; Not supported in MMS ROM
gp_con_fast_b:	equ	6		;
gp_dis_mtest:	equ	00100000b	; Not supported in MMS ROM
gp_dis_mtest_b:	equ	5		;
gp_boot_from_sec:	equ	00010000b	; boot from sec port (078h)
gp_boot_from_sec_b:	equ	4		; Not supported in MMS ROM

gp_pri_dev:	equ	00000011b
z89_pri_z17:	equ	00000000b
z89_pri_z47:	equ	00000001b
z89_pri_z67:	equ	00000010b
z89_pri_ndi:	equ	00000011b

gp_sec_dev:	equ	00001100b
z89_sec_z37:	equ	00000000b
z89_sec_z47:	equ	00000100b
z89_sec_z67:	equ	00001000b
z89_sec_ndi:	equ	00001100b

; H89 gp port latches (output)
gp_io1:		equ	10000000b	; 
gp_io0:		equ	01000000b	; 
gp_mem1:	equ	00100000b	; 
gp_mem0:	equ	00010000b	;
gp_mms2:	equ	00001000b	; MMS memory control
gp_mms1:	equ	00000100b	; MMS memory control
gp_2msen:	equ	00000010b	; 2ms "tic" interrupt enable
gp_ssen:	equ	00000001b	; single-step

; 8250 UART -----------------------------------------

; 8250 DLAB values for BAUD selection
dlab_9600:	equ	12
dlab_4800:	equ	24
dlab_2400:	equ	48
dlab_1200:	equ	96
dlab_600:	equ	192
dlab_300:	equ	384
dlab_110:	equ	1047

con_dat_port:	equ	con_base+i8250_data
con_dll_port:	equ	con_base+i8250_dll
con_ier_port:	equ	con_base+i8250_ier
con_iir_port:	equ	con_base+i8250_iir
con_dlm_port:	equ	con_base+i8250_dlm
con_fcr_port:	equ	con_base+i8250_fcr
con_lcr_port:	equ	con_base+i8250_lcr
con_mcr_port:	equ	con_base+i8250_mcr
con_lsr_port:	equ	con_base+i8250_lsr
con_msr_port:	equ	con_base+i8250_msr
con_scr_port:	equ	con_base+i8250_scr

prt_dat_port:	equ	prt_base+i8250_data
prt_dll_port:	equ	prt_base+i8250_dll
prt_ier_port:	equ	prt_base+i8250_ier
prt_iir_port:	equ	prt_base+i8250_iir
prt_dlm_port:	equ	prt_base+i8250_dlm
prt_fcr_port:	equ	prt_base+i8250_fcr
prt_lcr_port:	equ	prt_base+i8250_lcr
prt_mcr_port:	equ	prt_base+i8250_mcr
prt_lsr_port:	equ	prt_base+i8250_lsr
prt_msr_port:	equ	prt_base+i8250_msr
prt_scr_port:	equ	prt_base+i8250_scr

mdm_dat_port:	equ	mdm_base+i8250_data
mdm_dll_port:	equ	mdm_base+i8250_dll
mdm_ier_port:	equ	mdm_base+i8250_ier
mdm_iir_port:	equ	mdm_base+i8250_iir
mdm_dlm_port:	equ	mdm_base+i8250_dlm
mdm_fcr_port:	equ	mdm_base+i8250_fcr
mdm_lcr_port:	equ	mdm_base+i8250_lcr
mdm_mcr_port:	equ	mdm_base+i8250_mcr
mdm_lsr_port:	equ	mdm_base+i8250_lsr
mdm_msr_port:	equ	mdm_base+i8250_msr
mdm_scr_port:	equ	mdm_base+i8250_scr

alt_dat_port:	equ	alt_base+i8250_data
alt_dll_port:	equ	alt_base+i8250_dll
alt_ier_port:	equ	alt_base+i8250_ier
alt_iir_port:	equ	alt_base+i8250_iir
alt_dlm_port:	equ	alt_base+i8250_dlm
alt_fcr_port:	equ	alt_base+i8250_fcr
alt_lcr_port:	equ	alt_base+i8250_lcr
alt_mcr_port:	equ	alt_base+i8250_mcr
alt_lsr_port:	equ	alt_base+i8250_lsr
alt_msr_port:	equ	alt_base+i8250_msr
alt_scr_port:	equ	alt_base+i8250_scr

; Z37 Floppy Controller -----------------------------

z37_icl:	equ	z37_base+0
z37_acl:	equ	z37_base+1	; 0 -> cmd/data, 1 -> trk/sec
z37_sec:	equ	z37_base+2	; multiplexed by z37_acl w/z37_cmd
z37_trk:	equ	z37_base+3	; multiplexed by z37_acl w/z37_data
z37_cmd:	equ	z37_base+2
z37_data:	equ	z37_base+3

z37_icl_ds3:	equ	10000000b
z37_icl_ds2:	equ	01000000b
z37_icl_ds1:	equ	00100000b
z37_icl_ds0:	equ	00010000b
z37_icl_motor:	equ	00001000b
z37_icl_dden:	equ	00000100b
z37_icl_drqen:	equ	00000010b
z37_icl_inten:	equ	00000001b

; Z17 Hard-sectored Floppy --------------------------

; z17 is handled by legacy ROM... making calls into ROM.
z17_port0:	equ	z17_base+0
z17_port1:	equ	z17_base+1
z17_port2:	equ	z17_base+2
z17_port3:	equ	z17_base+3

; MMS 77314 IMI/Z47(REMEX) Interface ----------------

m314_imi_status:	equ	m314_base+0	; input
m314_imi_data:		equ	m314_base+1
m314_z47_data:		equ	m314_base+2
m314_z47_status:	equ	m314_base+3	; input
m314_z47_strobe:	equ	m314_base+3	; output

m314_z47_bsy:	equ	10000000b
m314_z47_dtr:	equ	01000000b
m314_z47_ddout:	equ	00100000b
m314_z47_err:	equ	00010000b
m314_z47_power:	equ	00000100b

m314_imi_power:	equ	10000000b
m314_imi_jumper:	equ	00000100b
m314_imi_ifact:	equ	00000010b
m314_imi_ready:	equ	00000001b

m314_gp_boot:	equ	11100000b	; field mask
m314_gp_imi:	equ	00000000b
m314_gp_1:	equ	00100000b
m314_gp_2:	equ	01000000b
m314_gp_3:	equ	01100000b
m314_gp_res4:	equ	10000000b
m314_gp_res5:	equ	10100000b
m314_gp_res6:	equ	11000000b
m314_gp_res7:	equ	11100000b
m314_gp_opts:	equ	00011100b	; field mask - specific 77316 unit

; MMS 77316 Floppy controller -----------------------

m316_data:	equ	m316_base+3
m316_trk:	equ	m316_base+wd1797_trk
m316_sec:	equ	m316_base+wd1797_sec
m316_cmd:	equ	m316_base+0

m316_gp_drv:	equ	00000111b	; drive select mask
m316_gp_5in:	equ	00000100b	; 5.25" drives are 4-7
m316_gp_ddn:	equ	01000000b
m316_gp_burstn:	equ	00100000b
m316_gp_ie:	equ	00001000b

; MMS 77320 SASI Interface --------------------------

m320_wr_data:	equ	m320_base+0
m320_control0:	equ	m320_base+1
m320_control1:	equ	m320_base+2
m320_rd_data:	equ	m320_base+0
m320_status:	equ	m320_base+1
m320_dipsw:	equ	m320_base+2

m320_sts_io:	equ	10000000b
m320_sts_req:	equ	01000000b
m320_sts_msg:	equ	00100000b
m320_sts_dc:	equ	00010000b
m320_sts_bsy:	equ	00001000b
m320_sts_2:	equ	00000100b
m320_sts_actint: equ	00000010b
m320_sts_ack:	equ	00000001b

m320_ct0_d7np:	equ	10000000b
m320_ct0_sel:	equ	01000000b
m320_ct0_ie:	equ	00100000b
m320_ct0_clr:	equ	00010000b

m320_ct1_dev0:	equ	00000000b
m320_ct1_dev1:	equ	00000001b
m320_ct1_dev2:	equ	00000010b
m320_ct1_dev3:	equ	00000011b
m320_ct1_dev4:	equ	00000100b
m320_ct1_dev5:	equ	00000101b
m320_ct1_dev6:	equ	00000110b
m320_ct1_dev7:	equ	00000111b

; Z67 Winchester Disk Interface ---------------------

; pri/sec base address... TBD
z67_data:	equ	z67_base1+0
z67_ctl:	equ	z67_base1+1
z67_dipsw:	equ	z67_base1+2

;z67_ctl_xxx:	equ	10000000b
z67_ctl_sel:	equ	01000000b
z67_ctl_inten:	equ	00100000b
z67_ctl_reset:	equ	00010000b
;z67_ctl_xxx:	equ	00001000b
;z67_ctl_xxx:	equ	00000100b
z67_ctl_daten:	equ	00000010b
;z67_ctl_xxx:	equ	00000001b

z67_sts_req:	equ	10000000b
z67_sts_inout:	equ	01000000b
z67_sts_msg:	equ	00100000b
z67_sts_cmd:	equ	00010000b
z67_sts_bsy:	equ	00001000b
z67_sts_pe:	equ	00000100b
z67_sts_irq:	equ	00000010b
z67_sts_ack:	equ	00000001b

; Z47 "Smart" 8" Drive Interface --------------------

z47_data:	equ	z47_base2+0
z47_status:	equ	z47_base2+1

z47_wsts_ie:	equ	10000000b
z47_wsts_reset:	equ	00000010b
z47_wsts_ffset:	equ	00000001b

z47_rsts_trn:	equ	10000000b
z47_rsts_ien:	equ	01000000b
z47_rsts_donen:	equ	00100000b
z47_rsts_swD:	equ	00010000b
z47_rsts_swC:	equ	00001000b
z47_rsts_swB:	equ	00000100b
z47_rsts_swA:	equ	00000010b
z47_rsts_errn:	equ	00000001b

; Unknown Device 37 ---------------------------------

x37_port0:	equ	x37_base+0
x37_port1:	equ	x37_base+1
x37_port2:	equ	x37_base+2
x37_port3:	equ	x37_base+3
x37_port4:	equ	x37_base+4
x37_port5:	equ	x37_base+5
x37_port6:	equ	x37_base+6	; data/command?

; Unknown Device 38 ---------------------------------

x38_port0:	equ	x38_base+0
x38_port1:	equ	x38_base+1
x38_port2:	equ	x38_base+2
x38_port3:	equ	x38_base+3
x38_port4:	equ	x38_base+4
x38_port5:	equ	x38_base+5	; device select

endif
